//**************************************************************************
//*                     This file is part of the                           *
//*                MMC - Mpxplay Multimedia Commander                      *
//*                   The source code of MMC is                            *
//*        (C) copyright 1998-2016 by PDSoft (Attila Padar)                *
//*                http://mpxplay.sourceforge.net                          *
//*                  email: mpxplay@freemail.hu                            *
//**************************************************************************
//*  This program is distributed in the hope that it will be useful,       *
//*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
//*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  *
//*  Please contact with the author (with me) if you want to use           *
//*  or modify this source.                                                *
//**************************************************************************
//function: ffmpeg video playing (main scheduler)

//#define MPXPLAY_USE_DEBUGF 1
#define DISPQT_DEBUGOUT_WARNING stderr
#define DISPQT_DEBUGOUT_QUEUE  stdout
#define DISPQT_DEBUGOUT_TIMED  NULL //stdout
#define DISPQT_DEBUGOUT_RENDER stdout
#define DISPQT_DEBUGOUT_SUBTITLE NULL // stdout
#define DISPQT_DEBUGOUT_VFCC stdout

#include "moc_video_qt.h"
#include "moc_mainwindow.h"

#define MPXPLAY_VIDEO_SCHEDULER_CPU2_AFFINITY_MASK ((mpxp_ptrsize_t)0x0001) // run scheduler (don't run decoder) thread on the 1. logical CPU, if there are more than 1 CPUs
#define MPXPLAY_VIDEO_SCHEDULER_CPU3_AFFINITY_MASK ((mpxp_ptrsize_t)0x0003) // run scheduler (don't run decoder) thread on the 1. and 2. logical CPUs, if there are more than 2 CPUs

#define INFFMPG_MUTEX_ONTIMER_TIMEOUT     10 // used on 10ms timer
#define MPXPLAY_VIDEO_WORKER_DRAWFRAME_RETRY 3

#define MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_DEFAULT    100
#define MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_SHORT      10
#define MPXPLAY_VIDEO_WORKER_THREAD_CRITICAL_ERROR_COUNT 10

#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG

extern "C" {
 extern unsigned long mpxplay_config_video_audiovisualization_type;
}

extern struct dispqt_videofilter_processdescriptor_s mpxplay_dispqt_videofilter_workerprocess_descriptors_all;
extern unsigned int mpxplay_dispqt_videofilter_workerprocess_nb_descriptors;

#if defined(PDS_THREADS_POSIX_THREAD)
static void *videoworker_video_worker_thread(void *decoder_data);
#else
static unsigned int videoworker_video_worker_thread(void *decoder_data);
#endif

FFMpegVideoDecoderWorker::FFMpegVideoDecoderWorker(MainWindow *main_window)
{
	this->main_window = main_window;
	pds_memset(&this->video_worker_infos, 0, sizeof(this->video_worker_infos));
//	this->video_worker_infos.main_window = main_window;
	pds_threads_mutex_new(&this->video_worker_infos.mutexhnd_workerinfo);
	this->videodworker_thread_id = pds_threads_thread_create((mpxp_thread_func)videoworker_video_worker_thread, (void *)&this->video_worker_infos, MPXPLAY_THREAD_PRIORITY_HIGHEST, 0);
#ifdef DISPQT_VIDEO_SCHEDULER_SEPARATED_AFFINITY
	// run video scheduler only on 1st Core (run other threads on the other CPUs)
	int nb_cpu = pds_threads_get_number_of_processors();
	mpxp_ptrsize_t cpu_affinity_scheduler = (nb_cpu >= 3)? MPXPLAY_VIDEO_SCHEDULER_CPU3_AFFINITY_MASK : ((nb_cpu >= 2)? MPXPLAY_VIDEO_SCHEDULER_CPU2_AFFINITY_MASK : 0); // TODO: check
	pds_threads_thread_set_affinity(this->videodworker_thread_id, cpu_affinity_scheduler);
#endif
}

FFMpegVideoDecoderWorker::~FFMpegVideoDecoderWorker()
{
	this->decoderworker_video_scheduler_control(DISPQT_VIDEOSCHEDULER_CONTROL_REQ_CLOSE);
	void *mutexhnd = this->video_worker_infos.mutexhnd_workerinfo;
	int lockerror = PDS_THREADS_MUTEX_LOCK(&mutexhnd, DISPQT_VIDEO_MUTEX_TIMEOUT);
	this->video_worker_infos.mutexhnd_workerinfo = NULL;
	this->cb_queue_clear(&this->video_worker_infos);
	pds_threads_thread_close(this->videodworker_thread_id);
	if(lockerror == MPXPLAY_ERROR_OK)
		PDS_THREADS_MUTEX_UNLOCK(&mutexhnd);
	pds_threads_mutex_del(&mutexhnd);
}

int FFMpegVideoDecoderWorker::cb_queue_put_entry(struct mpxplay_dispqt_ffmpegvideo_worker_info_s *workinfos, mpxplay_infile_callback_t *cbk_func, void *passdata, bool prepend)
{
	ffmpegvideo_callback_list_s *cbelem_new;

	cbelem_new = (ffmpegvideo_callback_list_s *)av_mallocz(sizeof(ffmpegvideo_callback_list_s));
	if(!cbelem_new)
		goto err_out_put_entry;

	cbelem_new->infile_passdata = passdata;
	funcbit_enable(cbelem_new->flags, (DISPQT_FFMPGVIDEOCALLBACKFLAG_FIRSTFRAME | DISPQT_FFMPGVIDEOCALLBACKFLAG_DISPLAYONEFRAME | DISPQT_FFMPGVIDEOCALLBACKFLAG_IGNORETIMESTAMP));
	if(prepend)
		funcbit_enable(cbelem_new->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_PRIMARYVIDEO);

	if(cbk_func)
	{
		mpxp_ptrsize_t cpu_affinity_decoder = 0;
#ifdef DISPQT_VIDEO_SCHEDULER_SEPARATED_AFFINITY
		int nb_cpu = pds_threads_get_number_of_processors();
		cpu_affinity_decoder = (nb_cpu >= 3)? ~MPXPLAY_VIDEO_SCHEDULER_CPU3_AFFINITY_MASK : ((nb_cpu >= 2)? ~MPXPLAY_VIDEO_SCHEDULER_CPU2_AFFINITY_MASK : 0);
#endif
		mpxplay_ffmpgdec_packetqueue_init(&cbelem_new->videoframe_queue, MPXPLAY_STREAMTYPEINDEX_VIDEO);
		cbelem_new->videodecoder_thread_id = pds_threads_thread_create((mpxp_thread_func)mpxplay_dispqt_video_decoder_worker_thread, cbelem_new, MPXPLAY_THREAD_PRIORITY_NORMAL, cpu_affinity_decoder);
		if(!cbelem_new->videodecoder_thread_id)
			goto err_out_put_entry;
	}

	cbelem_new->video_frame_filter_workerprocess = new DispQtVideoFrameFilter(this->main_window, this, &mpxplay_dispqt_videofilter_workerprocess_descriptors_all, mpxplay_dispqt_videofilter_workerprocess_nb_descriptors);
	if(!cbelem_new->video_frame_filter_workerprocess)
		goto err_out_put_entry;

	if (!workinfos->cblist_last){
		workinfos->cblist_first = cbelem_new;
		workinfos->cblist_last = cbelem_new;
	}else if(prepend){
		funcbit_disable(workinfos->cblist_first->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_PRIMARYVIDEO);
		cbelem_new->next = workinfos->cblist_first;
		workinfos->cblist_first = cbelem_new;
	}else{ // TODO: cb_queue_del_entry DISPQT_FFMPGVIDEOCALLBACKFLAG_PRIMARYVIDEO flag at append mode
		workinfos->cblist_last->next = cbelem_new;
		workinfos->cblist_last = cbelem_new;
	}
	cbelem_new->video_init_framecounter = DISPQT_FFMPGVIDEOCALLBACK_INITFRAMECOUNTER_START;
	cbelem_new->infile_callback = cbk_func;
	workinfos->nb_callbacks++;

	mpxplay_debugf(DISPQT_DEBUGOUT_QUEUE, "cb_queue_put n:%d f:%8.8X l:%8.8X cb:%8.8X pd:%8.8X", workinfos->nb_callbacks,
			(int)workinfos->cblist_first, (int)workinfos->cblist_last, (unsigned long)cbk_func, (unsigned long)passdata);

	return 0;

err_out_put_entry:
	this->cb_queue_elem_clear(cbelem_new);
	mpxplay_debugf(DISPQT_DEBUGOUT_QUEUE, "cb_queue_put FAILED");
	return -1;
}

void FFMpegVideoDecoderWorker::cb_queue_elem_clear(struct ffmpegvideo_callback_list_s *cbelem)
{
	if(!cbelem)
		return;
	mpxplay_debugf(DISPQT_DEBUGOUT_QUEUE, "cb_queue_elem_clear START pd:%8.8X", (unsigned int)cbelem->infile_passdata);
	if(cbelem->videodecoder_thread_id)
	{
		int retry = DISPQT_VIDEO_MUTEX_TIMEOUT / MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_SHORT;
		funcbit_enable(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_EXITEVENT);
		while(!funcbit_test(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_THEXITDONE) && (--retry)) { pds_threads_sleep(MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_SHORT); }
		pds_threads_thread_close(cbelem->videodecoder_thread_id);
	}
	cbelem->infile_callback = NULL;
	cbelem->infile_passdata = NULL;
	if(cbelem->videoframe_pktlist_elem)
		mpxplay_ffmpgdec_queuelistelem_clear(cbelem->videoframe_pktlist_elem);
	if(cbelem->video_pktlist_elem)
		mpxplay_ffmpgdec_queuelistelem_clear(cbelem->video_pktlist_elem);
	if(cbelem->subtitle_pktlist_elem)
		mpxplay_ffmpgdec_queuelistelem_clear(cbelem->subtitle_pktlist_elem);
	if(cbelem->subtitle_infos){
		mpxplay_video_ffmpeg_subtitleinfo_clear(cbelem->subtitle_infos);
		av_freep(&cbelem->subtitle_infos);
	}
	if(cbelem->vid_frame_decoded)
		av_frame_free(&cbelem->vid_frame_decoded);
	mpxplay_ffmpgdec_packetqueue_close(&cbelem->videoframe_queue);
	if(cbelem->video_frame_filter_workerprocess)
	{
		delete cbelem->video_frame_filter_workerprocess;
		cbelem->video_frame_filter_workerprocess = NULL;
	}
	av_free(cbelem);
	mpxplay_debugf(DISPQT_DEBUGOUT_QUEUE, "cb_queue_elem_clear END pd:%8.8X", (unsigned int)cbelem->infile_passdata);
}

int FFMpegVideoDecoderWorker::cb_queue_del_entry(struct mpxplay_dispqt_ffmpegvideo_worker_info_s *workinfos, mpxplay_infile_callback_t *cbk_func, void *passdata)
{
	ffmpegvideo_callback_list_s *cbelem = workinfos->cblist_first, *cbe_prev = NULL;

	do{
		if(!cbelem)
			return -1;
		if((cbelem->infile_callback == cbk_func) && (cbelem->infile_passdata == passdata)){
			if(cbe_prev)
				cbe_prev->next = cbelem->next;
			else if(cbelem == workinfos->cblist_first)
				workinfos->cblist_first = cbelem->next;
			if(cbelem == workinfos->cblist_last)
				workinfos->cblist_last = cbe_prev;
			cb_queue_elem_clear(cbelem);
			workinfos->nb_callbacks--;
			break;
		}
		cbe_prev = cbelem;
		cbelem = cbelem->next;
	}while(1);

	mpxplay_debugf(DISPQT_DEBUGOUT_QUEUE, "cb_queue_del n:%d f:%8.8X l:%8.8X cb:%8.8X pd:%8.8X", workinfos->nb_callbacks,
			(int)workinfos->cblist_first, (int)workinfos->cblist_last, (unsigned long)cbk_func, (unsigned long)passdata);

	return 0;
}

void FFMpegVideoDecoderWorker::cb_queue_clear(struct mpxplay_dispqt_ffmpegvideo_worker_info_s *workinfos)
{
	ffmpegvideo_callback_list_s *cbelem;
	if(!workinfos)
	{
		return;
	}
	cbelem = workinfos->cblist_first;
	while(cbelem){
		ffmpegvideo_callback_list_s *nextcbe = cbelem->next;
		cb_queue_elem_clear(cbelem);
		cbelem = nextcbe;
	}
	workinfos->cblist_first = NULL;
	workinfos->cblist_last = NULL;
	workinfos->nb_callbacks = 0;
}

static void videoworker_videosurface_is_open(struct mpxplay_dispqt_ffmpegvideo_worker_info_s *workinfos)
{
	bool is_videosurface_open = ((workinfos->nb_callbacks > 0) && (workinfos->is_video_stream || (mpxplay_config_video_audiovisualization_type != DISPQT_AUDIOVISUALIZATIONTYPE_NONE)))? true : false;
	if(is_videosurface_open != workinfos->is_videosurface_open)
	{
		if(!is_videosurface_open)
			workinfos->video_surface_refresh_request = DISPQT_VIDEOSURFACE_REFRESHTYPE_CLOSE;
		workinfos->is_videosurface_open = is_videosurface_open;
	}
}

bool FFMpegVideoDecoderWorker::videothread_ffmpeg_config_callback(unsigned int cfgcmd, mpxplay_infile_callback_t *cbk_func, void *passdata)
{
	struct mpxplay_dispqt_ffmpegvideo_worker_info_s *workinfos = &this->video_worker_infos;
	ffmpegvideo_callback_list_s *cbelem;
	mpxplay_packetlist_t *videoctx_pktelem;
	int video_streamtypeindex, lockerror;
	bool is_video_stream = false, is_videosurface_open;

	mpxplay_debugf(DISPQT_DEBUGOUT_VFCC, "videothread_ffmpeg_config_callback BEGIN cmd:%8.8X cbk:%8.8X pd:%8.8X", cfgcmd, (unsigned int)cbk_func, (unsigned int)passdata);

	if(cfgcmd == MPXPLAY_INFILE_CBKCFG_SRVFFMV_CLEAR_CALLBACKS)
	{
		workinfos->is_paused = true;
		workinfos->is_video_stream = false;
		workinfos->is_videosurface_open = false;
		return workinfos->is_videosurface_open;
	}
	else if((cfgcmd == MPXPLAY_INFILE_CBKCFG_SRVFFMV_PAUSE) || (cfgcmd == MPXPLAY_INFILE_CBKCFG_SRVFFMV_UNPAUSE) || (cfgcmd == MPXPLAY_INFILE_CBKCFG_SRVFFMV_RECHECK_SURFACE))
	{
		if(cfgcmd == MPXPLAY_INFILE_CBKCFG_SRVFFMV_PAUSE)
			workinfos->is_paused = true;
		else if(cfgcmd == MPXPLAY_INFILE_CBKCFG_SRVFFMV_UNPAUSE)
			workinfos->is_paused = false;
		videoworker_videosurface_is_open(workinfos);
		return workinfos->is_videosurface_open;
	}

	lockerror = PDS_THREADS_MUTEX_LOCK(&workinfos->mutexhnd_workerinfo, DISPQT_VIDEO_MUTEX_TIMEOUT);
	mpxplay_debugf(DISPQT_DEBUGOUT_VFCC, "videothread_ffmpeg_config_callback LOCKED cmd:%8.8X cbk:%8.8X pd:%8.8X lock:%d", cfgcmd, (unsigned int)cbk_func, (unsigned int)passdata, lockerror);

	switch(cfgcmd)
	{
//		case MPXPLAY_INFILE_CBKCFG_SRVFFMV_CLEAR_CALLBACKS: // FIXME: ?
//			this->cb_queue_clear(workinfos);
//			break;
		case MPXPLAY_INFILE_CBKCFG_SRVFFMV_OPEN_CALLBACK:
			this->cb_queue_put_entry(workinfos, cbk_func, passdata, true); // newly opened video is always the primary (putting it on the begin of cb queue)
			break;
		case MPXPLAY_INFILE_CBKCFG_SRVFFMV_CLOSE_CALLBACK:
			this->cb_queue_del_entry(workinfos, cbk_func, passdata);
			if(workinfos->nb_callbacks < 0)
				workinfos->nb_callbacks = 0;
			break;
		case MPXPLAY_INFILE_CBKCFG_SRVFFMV_REFRESH_CALLBACK:
			break;
	}

	if(workinfos->nb_callbacks > 0)
	{
		mpxplay_debugf(DISPQT_DEBUGOUT_VFCC, "videothread_ffmpeg_config_callback SEARCH cmd:%8.8X pd:%8.8X", cfgcmd, (unsigned int)passdata);
		cbelem = workinfos->cblist_first;
		while(cbelem)
		{
			int retval = MPXPLAY_ERROR_INFILE_NODATA;
			video_streamtypeindex = MPXPLAY_STREAMTYPEINDEX_VIDEO;
			videoctx_pktelem = NULL;
			if(cbelem->infile_callback)
				retval = cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_GET_CODECCTX_FFMPEG, (mpxp_ptrsize_t)&videoctx_pktelem, video_streamtypeindex, DISPQT_VIDEO_MUTEX_TIMEOUT);
			mpxplay_debugf(DISPQT_DEBUGOUT_VFCC, "videothread_ffmpeg_config_callback infile_callback FOUND ret:%d pd:%8.8X ctx:%8.8X next:%8.8X", retval,
							(int)cbelem->infile_passdata, (unsigned int)videoctx_pktelem, (unsigned int)cbelem->next);
			if((retval == MPXPLAY_ERROR_OK) && videoctx_pktelem)
			{
				is_video_stream = true;
			}
			//cbelem = cbelem->next;
			break; // currently we check the primary video only
		}
	}

	workinfos->is_video_stream = is_video_stream;
	videoworker_videosurface_is_open(workinfos);

	mpxplay_debugf(DISPQT_DEBUGOUT_VFCC, "videothread_ffmpeg_config_callback END open:%d nb:%d cmd:%8.8X pd:%8.8X lock:%d", (int)workinfos->is_videosurface_open, workinfos->nb_callbacks, cfgcmd, (int)passdata, lockerror);

	if(lockerror == MPXPLAY_ERROR_OK)
		PDS_THREADS_MUTEX_UNLOCK(&workinfos->mutexhnd_workerinfo);

	return workinfos->is_videosurface_open;
}

void FFMpegVideoDecoderWorker::videothread_ffmpeg_callback_call(int globalpos_x, int globalpos_y, unsigned long command, mpxp_ptrsize_t arg1, mpxp_ptrsize_t arg2)
{
	struct mpxplay_dispqt_ffmpegvideo_worker_info_s *workinfos = &this->video_worker_infos;
	if(workinfos->video_scheduler_control == DISPQT_VIDEOSCHEDULER_CONTROL_RUN)
	{
		mpxplay_debugf(DISPQT_DEBUGOUT_VFCC, "videothread_ffmpeg_callback_call BEGIN");
		int lockerror = PDS_THREADS_MUTEX_LOCK(&workinfos->mutexhnd_workerinfo, DISPQT_VIDEO_MUTEX_TIMEOUT);
		ffmpegvideo_callback_list_s *cbelem = workinfos->cblist_first; // TODO: select by globalpos
		if(cbelem && cbelem->infile_callback)
		{
			switch(command)
			{
				case MPXPLAY_INFILE_CBKCTRL_SET_SELECT_PROGRAM:
				case MPXPLAY_INFILE_CBKCTRL_SET_SELECT_STREAM:
					cbelem->callback_command.cmd = command;
					cbelem->callback_command.arg1 = arg1;
					cbelem->callback_command.arg2 = arg2;
					break;
				default:
					cbelem->infile_callback(cbelem->infile_passdata, command, arg1, arg2, DISPQT_VIDEO_MUTEX_TIMEOUT);
					break;
			}
		}
		if(lockerror == MPXPLAY_ERROR_OK)
			PDS_THREADS_MUTEX_UNLOCK(&workinfos->mutexhnd_workerinfo);
		mpxplay_debugf(DISPQT_DEBUGOUT_VFCC, "videothread_ffmpeg_callback_call END");
	}
}

bool FFMpegVideoDecoderWorker::decoderworker_video_scheduler_control(unsigned int control_req)
{
	struct mpxplay_dispqt_ffmpegvideo_worker_info_s *workinfos = &this->video_worker_infos;
	unsigned int expected_status = (control_req == DISPQT_VIDEOSCHEDULER_CONTROL_REQ_CLOSE)? DISPQT_VIDEOSCHEDULER_STATUS_CLOSED : DISPQT_VIDEOSCHEDULER_STATUS_SUSPENDED;
	mpxp_int64_t end_time = pds_gettimem() + DISPQT_VIDEO_MUTEX_TIMEOUT;
	bool success = false;

	do{
		if(control_req == DISPQT_VIDEOSCHEDULER_CONTROL_RUN)
		{
			success = true;
		}
		else if(workinfos->video_scheduler_control == control_req)
		{
			if(workinfos->video_scheduler_status == expected_status)
			{
				success = true;
			}
			else if(workinfos->video_scheduler_status != DISPQT_VIDEOSCHEDULER_CONTROL_RUN)
			{
				workinfos->video_scheduler_status = DISPQT_VIDEOSCHEDULER_CONTROL_RUN;
			}
		}
		workinfos->video_scheduler_control = control_req;
		if(success)
		{
			break;
		}
		pds_threads_sleep(MPXPLAY_THREADS_SHORTTASKSLEEP);
	}while(pds_gettimem() < end_time);

	if(control_req == DISPQT_VIDEOSCHEDULER_CONTROL_REQ_SUSPEND)
	{
		decoderworker_video_surface_refresh(DISPQT_VIDEOSURFACE_REFRESHTYPE_CLEAR);
	}

	return success;
}

void FFMpegVideoDecoderWorker::decoderworker_video_surface_refresh(unsigned int refresh_type)
{
	struct mpxplay_dispqt_ffmpegvideo_worker_info_s *workinfos = &this->video_worker_infos;
	if(refresh_type > workinfos->video_surface_refresh_request)
		workinfos->video_surface_refresh_request = refresh_type;
}

//------------------------------------------------------------------------------------------------------------------------------------------------------

// quality tune (reduce) based on video delay
static mpxp_bool_t ffmpegvideodecoderworker_video_delay_detect(ffmpegvideo_callback_list_s *cbelem, mpxp_int64_t video_delay, struct mmc_dispqt_config_s *gui_config)
{
	int framedrop_counter = DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYFRAMEDROP_NONE;
	mpxp_bool_t doFrameDrop = FALSE;

	if(cbelem->video_init_framecounter > 0)
	{
		cbelem->video_init_framecounter--;
		return doFrameDrop;
	}
	if(gui_config->selected_videoplayer_control & MPXPLAY_CONFIG_VIDEOPLAYERCONTROL_DISABLESKIPFRAME)
	{
		cbelem->video_delay_type = DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_NONE;
		cbelem->video_delay_large_counter = 0;
		cbelem->video_delay_framedrop_counter = 0;
		return doFrameDrop;
	}

	if(video_delay >= DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYLIMIT_MIDDLE)
	{
		if(cbelem->video_delay_large_counter >= DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYCOUNT_MIDDLE) // to avoid peek sync errors
		{
			if((video_delay >= DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYLIMIT_SIGNIFICANT) && (cbelem->video_delay_large_counter >= DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYCOUNT_SIGNIFICANT))
			{
				cbelem->video_delay_type = DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_SIGNIFICANT;
				framedrop_counter = DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYFRAMEDROP_SIGNIFICANT;
			}
			else if((video_delay >= DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYLIMIT_LARGE) && (cbelem->video_delay_large_counter >= DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYCOUNT_LARGE))
			{
				cbelem->video_delay_type = DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_LARGE;
				framedrop_counter = DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYFRAMEDROP_LARGE;
			}
			else
			{
				cbelem->video_delay_type = DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_MIDDLE;
			}
		}
		else
		{
			cbelem->video_delay_type = DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_SMALL;
		}
		if(cbelem->video_delay_large_counter < DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYCOUNT_SIGNIFICANT)
		{
			cbelem->video_delay_large_counter++;
		}
	}
	else
	{
		cbelem->video_delay_type = (video_delay >= DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYLIMIT_SMALL)? DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_SMALL : DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_NONE;
		if(cbelem->video_delay_large_counter > 0)
		{
			cbelem->video_delay_large_counter--;
		}
	}
	if((cbelem->video_delay_framedrop_counter > 0) && (framedrop_counter > 0))
	{
		cbelem->video_delay_framedrop_counter--;
		doFrameDrop = TRUE;
	}
	else
	{
		cbelem->video_delay_framedrop_counter = framedrop_counter;
	}
	return doFrameDrop;
}

static int videoworker_video_load_next_frame(ffmpegvideo_callback_list_s *cbelem, bool *resetvideo)
{
	int retval_video;
	if(cbelem->videoframe_pktlist_elem)
	{
		mpxplay_ffmpgdec_queuelistelem_clear(cbelem->videoframe_pktlist_elem);
		cbelem->videoframe_pktlist_elem = NULL;
	}
	retval_video = mpxplay_ffmpgdec_packetqueue_get(cbelem->infile_passdata, &cbelem->videoframe_queue, &cbelem->videoframe_pktlist_elem, 0);
	funcbit_disable(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_STILLPICT);
	if(cbelem->videoframe_pktlist_elem)
	{
		if(cbelem->videoframe_pktlist_elem->flags & MPXPLAY_PACKETLISTFLAG_CLEARBUF)
			*resetvideo = true;
		if((cbelem->videoframe_pktlist_elem->flags & MPXPLAY_PACKETLISTFLAG_STILLPICT) && !(cbelem->flags & DISPQT_FFMPGVIDEOCALLBACKFLAG_FIRSTFRAME))
			funcbit_enable(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_STILLPICT);
	}
	return retval_video;
}

static void videoworker_video_reset(struct mpxplay_dispqt_ffmpegvideo_worker_info_s *workinfos, ffmpegvideo_callback_list_s *cbelem, int video_index)
{
	if(workinfos->video_surface_refresh_request == DISPQT_VIDEOSURFACE_REFRESHTYPE_NONE)
		workinfos->video_surface_refresh_request = DISPQT_VIDEOSURFACE_REFRESHTYPE_RESET;

	mpxplay_dispqt_ffmpegvideo_frame_display(video_index, cbelem->flags, NULL, NULL, workinfos->video_surface_refresh_request);

	workinfos->video_surface_refresh_request = DISPQT_VIDEOSURFACE_REFRESHTYPE_NONE;
	if(funcbit_test(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_FIRSTFRAME))
	{
		cbelem->video_delay_type = DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_NONE;
		cbelem->video_delay_large_counter = 0;
		funcbit_disable(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_FIRSTFRAME);
	}
	cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_SET_PACKETQUEUE_CLEAR, MPXPLAY_STREAMTYPEINDEX_SUBTITLE, MPXPLAY_PACKETLISTFLAG_CLEARBUF, DISPQT_MUTEX_TIMEOUT);
	mpxplay_decoderworker_subtitlepkt_clear(cbelem);
	mpxplay_video_ffmpeg_subtitleinfo_clear(cbelem->subtitle_infos);
	cbelem->video_last_timestamp = 0;
	cbelem->subtitle_begin_timestamp = cbelem->subtitle_end_timestamp = 0;
	funcbit_enable(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_IGNORETIMESTAMP);
	if(workinfos->is_paused)
		funcbit_enable(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_DISPLAYONEFRAME);
	mpxplay_debugf(DISPQT_DEBUGOUT_WARNING, "TimedEvent RESETVIDEO");
}

static bool videoworker_video_frame_scheduler(struct mpxplay_dispqt_ffmpegvideo_worker_info_s *workinfos, ffmpegvideo_callback_list_s *cbelem, int video_index, mpxp_int64_t timestamp_curr_us)
{
	bool videoframe_rendered = false;
	if(cbelem->videoframe_pktlist_elem)
	{
		mpxp_int64_t timestamp_vpacket_us = cbelem->videoframe_pktlist_elem->timestamp_us;
		mpxp_int64_t video_delay = timestamp_curr_us - timestamp_vpacket_us;
		mpxp_int64_t curr_time_ms = pds_gettimem(), render_time_diff_ms = curr_time_ms - cbelem->video_last_render_time;
		mpxp_bool_t doIgnoreTimestamp = (funcbit_test(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_IGNORETIMESTAMP))? TRUE : FALSE;
#ifdef DISPQT_FFMPEG_TEST_SPEED
		mpxp_int64_t timedevent_start_time_ms = pds_gettimem(), gap_timedevent_ms = timedevent_start_time_ms - workinfos->timedevent_prev_start_time;
		workinfos->timedevent_prev_start_time = timedevent_start_time_ms;
//		mpxplay_debugf(DISPQT_DEBUGOUT_RENDER, "TimedEvent TIMESYNC cts:%lld tcu:%lld vts:%lld cd:%5lld vd:%lld", curr_time_ms, timestamp_curr_us, timestamp_vpacket_us, (timestamp_curr_us - cbelem->timedevent_prev_sync_time), video_delay);
#endif
		if(!doIgnoreTimestamp && (timestamp_curr_us > 0) && (cbelem->video_last_timestamp > 0))
		{
			if(timestamp_vpacket_us > (timestamp_curr_us + DISPQT_FFMPGVIDEO_AVSYNC_MAX_VALID_DIFF_US))
			{
				mpxplay_debugf(DISPQT_DEBUGOUT_TIMED, "TimedEvent vd:%lld rtd:%lld", (timestamp_vpacket_us - cbelem->video_last_timestamp), render_time_diff_ms);
				if((timestamp_vpacket_us - cbelem->video_last_timestamp) < (render_time_diff_ms * 1000))
				{
					doIgnoreTimestamp = TRUE; // to start video with invalid PTS too
				}
			}
		}

		if( doIgnoreTimestamp
		 || ((timestamp_curr_us > 0) && (timestamp_vpacket_us <= timestamp_curr_us))
		 || (cbelem->video_last_timestamp && (
				    (timestamp_vpacket_us < cbelem->video_last_timestamp)
				 || (timestamp_vpacket_us > (cbelem->video_last_timestamp + DISPQT_FFMPGVIDEO_VIDEO_MAX_FRAME_DIFF_US)))
			)
		){
			ffmpegvideo_subtitle_info_s *subtitle = mpxplay_decoderworker_subtitle_get(cbelem, timestamp_curr_us);
			AVFrame *videoout_frame = (AVFrame *)(cbelem->videoframe_pktlist_elem->frame_pkt);
			mpxp_bool_t doFrameDrop = FALSE;
#ifdef DISPQT_FFMPEG_TEST_SPEED
			mpxp_int64_t render_time_length = curr_time_ms, sync_diff = (timestamp_curr_us - cbelem->timedevent_prev_sync_time) / 1000;
			cbelem->timedevent_prev_sync_time = timestamp_curr_us;
//			mpxplay_debugf(DISPQT_DEBUGOUT_TIMED, "TimedEvent VIDEOOUT n:%d vpts:%lld sync:%lld SDIFF:%2lld VDIFF:%lld PDIFF:%lld", cbelem->videoframe_queue.nb_packets, timestamp_vpacket_us, timestamp_curr_us, sync_diff, (timestamp_vpacket_us - cbelem->video_last_timestamp), video_delay);
#endif
			if(!funcbit_test(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_IGNORETIMESTAMP) && !(videoout_frame->flags & DISPQT_AVFRAME_FLAG_NOSKIPFRAME))
			{
				doFrameDrop = ffmpegvideodecoderworker_video_delay_detect(cbelem, video_delay, &mmc_config);
			}

			if(  funcbit_test(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_IGNORETIMESTAMP)
			 || (render_time_diff_ms >= INFFMPEG_VIDEO_RENDER_MAX_INTERVAL_MS)
			 || ((doFrameDrop == FALSE) && (render_time_diff_ms >= INFFMPEG_VIDEO_RENDER_MIN_INTERVAL_MS))
			){
				if(cbelem->videoframe_pktlist_elem->flags & MPXPLAY_PACKETLISTFLAG_STILLPICT)  // TODO: check: still pictures always use pixel aspect ratio override
					videoout_frame->sample_aspect_ratio.num = videoout_frame->sample_aspect_ratio.den = 1;

				mpxplay_dispqt_ffmpegvideo_frame_display(video_index, cbelem->flags, videoout_frame, subtitle, workinfos->video_surface_refresh_request);

				videoframe_rendered = true;
				funcbit_disable(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_FIRSTFRAME);
				cbelem->video_last_render_time = curr_time_ms;
				cbelem->videoframe_pktlist_elem->frame_pkt = NULL;  // to keep AVFrame for videowidget_ffmpeg_frame_display
#ifdef DISPQT_FFMPEG_TEST_SPEED
				render_time_length = pds_gettimem() - render_time_length;
//				if((render_time_diff_ms < 35) || (render_time_diff_ms > 50) || (sync_diff < 35) || (sync_diff > 55) || (render_time_length >= (DISPQT_FFMPEG_WORKERTHREAD_REFRESH_MS * 2))) // 25 fps
//				if((render_time_diff_ms < 30) || (render_time_diff_ms > 40) || (sync_diff < 28) || (sync_diff > 45) || (render_time_length >= (DISPQT_FFMPEG_WORKERTHREAD_REFRESH_MS * 2))) // 30 fps
//				if((video_delay > 10000) || (render_time_length >= (DISPQT_FFMPEG_WORKERTHREAD_REFRESH_MS * 2)))
				{
					mpxplay_debugf(DISPQT_DEBUGOUT_RENDER, "GAP rtd:%3d rtl:%2llu gte:%2llu sd:%2lld vd:%lld vy:%2lld vdt:%d lc:%d nb:%d",
						(int)render_time_diff_ms, render_time_length, gap_timedevent_ms, sync_diff, (timestamp_vpacket_us - cbelem->video_last_timestamp + 500) / 1000,
						(video_delay + 500 ) / 1000, cbelem->video_delay_type, cbelem->video_delay_large_counter, cbelem->videoframe_queue.nb_packets);
				}
#endif
			}
			else
			{
#ifdef DISPQT_FFMPEG_TEST_SPEED
				mpxplay_debugf(DISPQT_DEBUGOUT_RENDER, "DRP rtd:%3d sd:%lld vd:%lld vy:%lld vdt:%d dc:%d", (int)render_time_diff_ms, sync_diff, ((timestamp_vpacket_us - cbelem->video_last_timestamp) / 1000), (video_delay / 1000), cbelem->video_delay_type, cbelem->video_delay_framedrop_counter);
#else
				mpxplay_debugf(DISPQT_DEBUGOUT_RENDER, "DRP rtd:%3d vd:%lld vy:%lld vdt:%d", (int)render_time_diff_ms, ((timestamp_vpacket_us - cbelem->video_last_timestamp) / 1000), (video_delay / 1000), cbelem->video_delay_type);
#endif
			}
			cbelem->video_last_timestamp = timestamp_vpacket_us;
			mpxplay_ffmpgdec_queuelistelem_clear(cbelem->videoframe_pktlist_elem);
			cbelem->videoframe_pktlist_elem = NULL;
			funcbit_disable(cbelem->flags, (DISPQT_FFMPGVIDEOCALLBACKFLAG_DISPLAYONEFRAME | DISPQT_FFMPGVIDEOCALLBACKFLAG_IGNORETIMESTAMP));
		}
		else if((video_delay < 0) && (cbelem->video_delay_type != DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_NONE))
		{
			ffmpegvideodecoderworker_video_delay_detect(cbelem, video_delay, &mmc_config);
		}
	}
	return videoframe_rendered;
}

#if defined(PDS_THREADS_POSIX_THREAD)
static void *videoworker_video_worker_thread(void *decoder_data)
#else
static unsigned int videoworker_video_worker_thread(void *decoder_data)
#endif
{
	struct mpxplay_dispqt_ffmpegvideo_worker_info_s *workinfos = (struct mpxplay_dispqt_ffmpegvideo_worker_info_s *)decoder_data;

	do{
		int video_index = 0, sleep_length = 5, retry;
		ffmpegvideo_callback_list_s *cbelem;
		bool video_frame_rendered = false;

		if(!workinfos || !workinfos->mutexhnd_workerinfo)
			break;
		if(workinfos->video_scheduler_control != DISPQT_VIDEOSCHEDULER_CONTROL_RUN)
		{
			if(workinfos->video_scheduler_control == DISPQT_VIDEOSCHEDULER_CONTROL_REQ_CLOSE)
			{
				break;
			}
			workinfos->video_scheduler_status = DISPQT_VIDEOSCHEDULER_STATUS_SUSPENDED;
			//if(workinfos->video_surface_refresh_request >= DISPQT_VIDEOSURFACE_REFRESHTYPE_CLEAR)
			//	goto err_out_refresh;
			goto do_sleep_thread;
		}

		cbelem = workinfos->cblist_first;
		while(cbelem)
		{
			ffmpegvideo_callback_command_s *ccmd = &cbelem->callback_command;
			if(ccmd->cmd)
			{
				if(cbelem->infile_callback)
					cbelem->infile_callback(cbelem->infile_passdata, ccmd->cmd, ccmd->arg1, ccmd->arg2, DISPQT_VIDEO_MUTEX_TIMEOUT);
				ccmd->cmd = 0;
			}
			cbelem = cbelem->next;
		}

		workinfos->video_scheduler_status = DISPQT_VIDEOSCHEDULER_CONTROL_RUN;

		if(!workinfos->is_videosurface_open)
		{
			goto err_out_refresh;
		}

		mpxplay_debugf(DISPQT_DEBUGOUT_TIMED, "decoderworker_TimedEvent BEGIN");

		retry = MPXPLAY_VIDEO_WORKER_DRAWFRAME_RETRY;
		cbelem = workinfos->cblist_first;
		while(cbelem)
		{
			int retval_video = -1;
//			mpxplay_debugf(DISPQT_DEBUGOUT_TIMED, "TimedEvent BEGIN n:%d f:%8.8X l:%8.8X cb:%8.8X pd:%8.8X qn:%d", workinfos->nb_callbacks,
//				(int)workinfos->cblist_first, (int)workinfos->cblist_last, (unsigned long)workinfos->cblist_first->infile_callback, 
//				(unsigned long)workinfos->cblist_first->infile_passdata, cbelem->videoframe_queue.nb_packets);
			if(cbelem->infile_callback && cbelem->infile_passdata)
			{
				bool resetvideo = (mpxplay_ffmpgdec_packetqueue_clear(&cbelem->videoframe_queue, MPXPLAY_PACKETLISTFLAG_CLEARBUF))? true : false;
				if(!cbelem->videoframe_pktlist_elem || resetvideo)
				{
					retval_video = videoworker_video_load_next_frame(cbelem, &resetvideo);
					if(resetvideo && (cbelem == workinfos->cblist_first) && funcbit_test(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_PRIMARYVIDEO))
					{
						videoworker_video_reset(workinfos, cbelem, video_index);
					}
				}
				if(workinfos->is_paused && !funcbit_test(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_DISPLAYONEFRAME))
				{
//					mpxplay_debugf(DISPQT_DEBUGOUT_TIMED, "TimedEvent PAUSED");
					break;
				}
				if(!cbelem->subtitle_pktlist_elem)
				{
					mpxplay_decoderworker_subtitle_load_next_frame(cbelem);
				}
//				mpxplay_debugf(DISPQT_DEBUGOUT_TIMED, "infile_callback END pd:%8.8X r:%d vf:%8.8X", (unsigned int)cbelem->infile_passdata, retval_video, (unsigned int)cbelem->videoframe_pktlist_elem);
			}
			if((cbelem == workinfos->cblist_first) && funcbit_test(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_PRIMARYVIDEO))
			{
				mpxp_int64_t timestamp_curr_us = (cbelem->videoframe_pktlist_elem)? cbelem->videoframe_pktlist_elem->timestamp_us : 0;
				if(cbelem->videoframe_pktlist_elem || cbelem->subtitle_end_timestamp)
					cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_GET_TIMESYNC, (mpxp_ptrsize_t)&timestamp_curr_us, MPXPLAY_STREAMTYPEINDEX_VIDEO, INFFMPG_MUTEX_ONTIMER_TIMEOUT);
				video_frame_rendered |= videoworker_video_frame_scheduler(workinfos, cbelem, video_index, timestamp_curr_us);
				mpxplay_decoderworker_subtitle_decode(cbelem, timestamp_curr_us);
			}
			else // TODO: currently we simply drop non primary video packets (later can be video crossfade or ...)
			{
				mpxplay_decoderworker_subtitlepkt_clear(cbelem);
			}
			if(!cbelem->videoframe_pktlist_elem && (retval_video >= 0) && (--retry))
				continue;
			retry = MPXPLAY_VIDEO_WORKER_DRAWFRAME_RETRY;
//			mpxplay_debugf(DISPQT_DEBUGOUT_TIMED, "TimedEvent END n:%d f:%8.8X l:%8.8X cb:%8.8X pd:%8.8X", workinfos->nb_callbacks,
//							(int)workinfos->cblist_first, (int)workinfos->cblist_last, (unsigned long)workinfos->cblist_first->infile_callback, 
//							(unsigned long)workinfos->cblist_first->infile_passdata);
			cbelem = cbelem->next;
			video_index++;
		}

err_out_refresh:
		if(!video_frame_rendered && (workinfos->video_surface_refresh_request != DISPQT_VIDEOSURFACE_REFRESHTYPE_NONE))
		{
			unsigned int refresh_type = workinfos->video_surface_refresh_request;
			workinfos->video_surface_refresh_request = DISPQT_VIDEOSURFACE_REFRESHTYPE_NONE;
			if(mpxplay_dispqt_ffmpegvideo_frame_display(0, 0, NULL, NULL, refresh_type) != 0) // DISPQT_ERROR_OK  - initial clear delay
			{
				workinfos->video_surface_refresh_request = refresh_type;
			}
		}
do_sleep_thread:
		pds_threads_sleep(sleep_length);
	}while(workinfos->video_scheduler_control != DISPQT_VIDEOSCHEDULER_CONTROL_REQ_CLOSE);
	workinfos->video_scheduler_status = DISPQT_VIDEOSCHEDULER_STATUS_CLOSED;
	pds_threads_sleep(DISPQT_VIDEO_MUTEX_TIMEOUT);  // required for pds_threads_thread_close -> TerminateThread
	return 0;
}

#endif // MPXPLAY_LINK_ORIGINAL_FFMPEG
